/*
 * Copyright (c) 2025      Andreas Schneider <asn@samba.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

use core::ffi::c_char;
use std::collections::HashMap;
use std::ffi::CString;
use std::ptr;

use crate::{
    AutoResizeMode, Fido2HmacSalt, Pkcs11EncryptedKey, RecoveryKey, UserBinding, UserDisposition,
    UserHomeStorage, UserPerMachine, UserPrivileged, UserRecord, UserSecret, UserSignature,
    UserStatus,
};

use crate::c_utils::{
    KirmesStrVec, KirmesStrVecMap, c_string_free, kirmes_boxed_vec_free, kirmes_str_vec_free,
    option_string_to_c_char, string_to_c_char,
};

macro_rules! box_free {
    ($input:expr) => {{
        if !$input.is_null() {
            unsafe {
                let _ = Box::from_raw($input.cast_mut());
            }
        }
    }};
}

/// User Records
///
/// For a detailed explanation please visit <https://systemd.io/USER_RECORDS>
#[repr(C)]
#[allow(clippy::struct_excessive_bools)]
#[derive(Debug)]
pub struct KirmesUserRecord {
    //
    // Regular section
    //
    /// The UNIX user name for this record.
    pub user_name: *const c_char,
    /// The 'realm' a user is defined in.
    pub realm: *const c_char,
    /// The absolute path to a world-readable copy of the user’s blob directory.
    pub blob_directory: *const c_char,
    /// An object, which maps valid blob directory filenames to SHA256 hashes formatted as hex
    /// strings.
    pub blob_manifest: KirmesStrVecMap,
    /// The real name of the user.
    pub real_name: *const c_char,
    /// The email address of the user.
    pub email_address: *const c_char,
    /// The name of an icon picked by the user.
    pub icon_name: *const c_char,
    /// A free-form location string describing the location of the user.
    pub location: *const c_char,
    /// Clarifies the disposition of the user, i.e. the context it is defined in.
    pub disposition: KirmesUserDisposition,
    /// A timestamp indicating when the user record was last changed (µs since the epoch 1970).
    pub last_change_usec: u64,
    /// A timestamp indicating the point in time when password of the user was last changed (µs
    /// since the epoch 1970).
    pub last_password_change_usec: u64,
    /// The shell binary to use for terminal logins of this user.
    pub shell: *const c_char,
    /// The `umask` to set for the user’s login sessions.
    pub umask: u32,
    /// An array of strings, each containing an environment variable and its value to set for the
    /// user’s login session.
    pub environment: KirmesStrVecMap,
    /// The preferred timezone of the user.
    pub time_zone: *const c_char,
    /// The preferred language/locale for the user.
    pub preferred_language: *const c_char,
    /// The preferred languages/locales that should be used in the event that translations for the
    /// `preferredLanguage` are missing
    pub additional_languages: KirmesStrVec,
    /// An integer value in the range -20…19
    pub nice_level: i32,
    // An object, where each key refers to a Linux resource limit (such as RLIMIT_NOFILE and
    // similar).
    // TODO pub resource_limits: ???,
    /// If true, the user account is locked, the user may not log in.
    pub locked: bool,
    /// A timestamp when record should be considered invalid for the purpose of logging in (µs
    /// since the UNIX epoch 1970).
    pub not_before_usec: u64,
    /// The point in time after which logins shall not be permitted anymore (µs since the
    /// UNIX epoch 1970).
    pub not_after_usec: u64,
    /// The storage mechanism for the user’s home directory.
    pub storage: KirmesUserHomeStorage,
    /// The intended home directory disk space in bytes to assign to the user.
    pub disk_size: u64,
    /// Similar to diskSize but takes a relative value, but specifies a fraction of the available
    /// disk space on the selected storage medium to assign to the user.
    pub disk_size_relative: u64,
    /// The absolute path to the skeleton directory to populate a new home directory from.
    pub skeleton_directory: *const c_char,
    /// The UNIX access mask in the range 0…511 for the home directory when it is first created.
    pub access_mask: u32,
    /// The maximum number of tasks the user may start in parallel during system runtime.
    pub tasks_max: u64,
    /// Upper memory limits for all processes of the user.
    pub memory_high: u64,
    /// Upper memory limits for all processes of the user.
    pub memory_max: u64,
    /// CPU scheduling weights for the user’s processes (range 1…10000 ,defaults to
    /// 100).
    pub cpu_weight: u32,
    /// IO scheduling weights for the user’s processes (range 1…10000 ,defaults to 100).
    pub io_weight: u32,
    /// Control nosuid mount flag of the user’s home directories.
    pub mount_no_suid: bool,
    /// Control nodev mount flag of the user’s home directories.
    pub mount_no_dev: bool,
    /// Control noexec mount flag of the user’s home directories.
    pub mount_no_exec: bool,
    /// The Windows File Sharing domain (CIFS) to use.
    pub cifs_domain: *const c_char,
    /// The Windows File Sharing user name (CIFS) to associate this user record with.
    pub cifs_user_name: *const c_char,
    /// The Windows File Share service (CIFS) to mount as home directory of the user on login.
    pub cifs_service: *const c_char,
    /// Additional mount options to pass to mount.cifs when mounting the home directory CIFS share.
    pub cifs_extra_mount_options: *const c_char,
    /// An absolute file system path to the file, directory or block device to use for storage
    /// backing the home directory.
    pub image_path: *const c_char,
    /// An absolute file system path to the home directory.
    pub home_directory: *const c_char,
    /// The numeric UNIX user ID (UID) to use for the user.
    pub uid: u32,
    /// The numeric UNIX group ID (GID) to use for the user.
    pub gid: u32,
    /// An array of strings, each indicating a UNIX group this user shall be a member of.
    pub member_of: KirmesStrVec,
    /// One of ext4, xfs, btrfs (possibly others) to use as file system for the user’s home
    /// directory.
    pub file_system_type: *const c_char,
    /// A lower-case, text-formatted UUID, referencing the GPT partition UUID the home directory is
    /// located in.
    pub partition_uuid: *const c_char,
    /// A lower-case, text-formatted UUID, referencing the LUKS volume UUID the home directory is
    /// located in.
    pub luks_uuid: *const c_char,
    /// A lower-case, text-formatted UUID, referencing the file system UUID the home directory is
    /// located in.
    pub file_system_uuid: *const c_char,
    ///  If true and luks storage is used, controls whether the loopback block devices, LUKS and
    ///  the file system on top shall be used in discard mode.
    pub luks_discard: bool,
    /// Similar to luksDiscard, it controls whether to trim/allocate the file system/backing file
    /// when deactivating the home directory.
    pub luks_offline_discard: bool,
    /// Additional mount options to append to the default mount options for the file system in the
    /// LUKS volume.
    pub luks_extra_mount_options: *const c_char,
    /// The cipher to use for the LUKS storage mechanism.
    pub luks_cipher: *const c_char,
    // Selects the cipher mode to use for the LUKS storage mechanism.
    pub luks_cipher_mode: *const c_char,
    /// The volume key length in bytes to use for the LUKS storage mechanism.
    pub luks_volume_key_size: u64,
    /// Selects the hash algorithm to use for the PBKDF operation.
    pub luks_pbkdf_hash_algorithm: *const c_char,
    /// The PBKDF type to use for the LUKS storage mechanism.
    pub luks_pbkdf_type: *const c_char,
    /// The intended number of iterations for the PBKDF operation.
    pub luks_pbkdf_force_iterations: u64,
    /// The intended time cost for the PBKDF operation, in µs.
    pub luks_pbkdf_time_cost_usec: u64,
    /// The intended memory cost for the PBKDF operation, in bytes.
    pub luks_pbkdf_memory_cost: u64,
    /// The intended required parallel threads for the PBKDF operation.
    pub luks_pbkdf_parallel_threads: u64,
    /// The sector size to use for the LUKS storage mechanism, in bytes.
    pub luks_sector_size: u64,
    /// Controls whether the home area shall be grown automatically.
    pub auto_resize_mode: KirmesAutoResizeMode,
    /// Configures the free disk space rebalancing weight for the home area.
    pub rebalance_weight: u64,
    /// The service that defines or manages this user record.
    pub service: *const c_char,
    /// Configures the authentication rate limiting enforced on the user account.
    pub rate_limit_interval_usec: u64,
    /// Closely related to rateLimitIntervalUSec, that puts a limit on authentication attempts
    /// within the configured time interval.
    pub rate_limit_interval_burst: u64,
    /// Configures whether to enforce the system’s password policy when creating the home directory
    /// for the user or changing the user’s password.
    pub enforce_password_policy: bool,
    /// If true the user record is marked as suitable for auto-login.
    pub auto_login: bool,
    /// The user’s preferred session type (i.e. x11, wayland, or other values valid for
    /// `$XDG_SESSION_TYPE`).
    pub preferred_session_type: *const c_char,
    /// The user’s preferred session launcher desktop entry file (i.e. gnome, plasma, ...)
    pub preferred_session_launcher: *const c_char,
    /// The time in µs the per-user service manager is kept around after the user fully logged out.
    pub stop_delay_usec: u64,
    ///  If true all processes of the user are automatically killed when the user logs out.
    pub kill_processes: bool,
    /// Encoding how much time has to pass at least between password changes of the user.
    pub password_change_min_usec: u64,
    /// Encodes how much time has to pass at at most between password changes of the user.
    pub password_change_max_usec: u64,
    /// How much time to warn the user before their password expires, in µs.
    pub password_change_warn_usec: u64,
    /// Encodes how much time has to pass after the password expired that the account is
    /// deactivated, in µs.
    pub password_change_inactive_usec: u64,
    /// If true the user has to change their password on next login.
    pub password_change_now: bool,
    /// An array of strings, each with an RFC 7512 compliant PKCS#11 URI referring to security
    /// token (or smart card) of some form, that shall be associated with the user and may be used
    /// for authentication.
    pub pkcs11_token_uri: KirmesStrVec,
    /// An array of strings, each with a Base64-encoded FIDO2 credential ID that shall be used for
    /// authentication with FIDO2 devices that implement the hmac-secret extension.
    pub fido2_hmac_credential: KirmesStrVec,
    /// An array of strings, each indicating the type of one recovery key.
    pub recovery_key_type: KirmesStrVec,
    /// An array of strings, each corresponding to a field name that can appear in the regular or
    /// perMachine sections.
    pub self_modifiable_fields: KirmesStrVec,
    /// Similar to selfModifiableFields, but it lists blobs that the user is allowed to edit.
    pub self_modifiable_blobs: KirmesStrVec,
    /// Similar to selfModifiableFields, but it lists fields in the privileged section that the
    /// user is allowed to edit.
    pub self_modifiable_privileged: KirmesStrVec,

    //
    // Privileged section
    //
    pub privileged: *const KirmesUserPrivileged,

    //
    // PerMachine section (Array)
    //
    pub per_machine: KirmesUserPerMachineVec,

    //
    // Binding section (Array)
    //
    pub binding: KirmesUserBindingVec,

    //
    // Status section (Array)
    //
    pub status: KirmesUserStatusVec,

    //
    // Signature section (Array)
    //
    pub signature: KirmesUserSignatureVec,

    //
    // Secret section
    //
    pub secret: *const KirmesUserSecret,
}

impl From<UserRecord> for KirmesUserRecord {
    #[allow(clippy::too_many_lines)]
    fn from(u: UserRecord) -> Self {
        KirmesUserRecord {
            user_name: string_to_c_char(u.user_name),
            realm: option_string_to_c_char(u.realm),
            blob_directory: option_string_to_c_char(u.blob_directory),
            blob_manifest: u.blob_manifest.into(),
            real_name: option_string_to_c_char(u.real_name),
            email_address: option_string_to_c_char(u.email_address),
            icon_name: option_string_to_c_char(u.icon_name),
            location: option_string_to_c_char(u.location),
            disposition: u.disposition.into(),
            last_change_usec: u.last_change_usec.unwrap_or(u64::MAX),
            last_password_change_usec: u.last_password_change_usec.unwrap_or(u64::MAX),
            shell: option_string_to_c_char(u.shell),
            umask: u.umask.unwrap_or(u32::MAX),
            environment: u.environment.into(),
            time_zone: option_string_to_c_char(u.time_zone),
            preferred_language: option_string_to_c_char(u.preferred_language),
            additional_languages: u.additional_languages.into(),
            nice_level: u.nice_level.unwrap_or(i32::MAX),
            locked: u.locked,
            not_before_usec: u.not_before_usec.unwrap_or(u64::MAX),
            not_after_usec: u.not_after_usec.unwrap_or(u64::MAX),
            storage: u.storage.into(),
            disk_size: u.disk_size.unwrap_or(u64::MAX),
            disk_size_relative: u.disk_size_relative.unwrap_or(u64::MAX),
            skeleton_directory: option_string_to_c_char(u.skeleton_directory),
            access_mask: u.access_mask.unwrap_or(u32::MAX),
            tasks_max: u.tasks_max.unwrap_or(u64::MAX),
            memory_high: u.memory_high.unwrap_or(u64::MAX),
            memory_max: u.memory_max.unwrap_or(u64::MAX),
            cpu_weight: u.cpu_weight.unwrap_or(u32::MAX),
            io_weight: u.io_weight.unwrap_or(u32::MAX),
            mount_no_suid: u.mount_no_suid,
            mount_no_dev: u.mount_no_dev,
            mount_no_exec: u.mount_no_exec,
            cifs_domain: option_string_to_c_char(u.cifs_domain),
            cifs_user_name: option_string_to_c_char(u.cifs_user_name),
            cifs_service: option_string_to_c_char(u.cifs_service),
            cifs_extra_mount_options: option_string_to_c_char(u.cifs_extra_mount_options),
            image_path: option_string_to_c_char(u.image_path),
            home_directory: option_string_to_c_char(u.home_directory),
            uid: u.uid.unwrap_or(u32::MAX),
            gid: u.gid.unwrap_or(u32::MAX),
            member_of: u.member_of.into(),
            file_system_type: option_string_to_c_char(u.file_system_type),
            partition_uuid: option_string_to_c_char(u.partition_uuid),
            luks_uuid: option_string_to_c_char(u.luks_uuid),
            file_system_uuid: option_string_to_c_char(u.file_system_uuid),
            luks_discard: u.luks_discard,
            luks_offline_discard: u.luks_offline_discard,
            luks_extra_mount_options: option_string_to_c_char(u.luks_extra_mount_options),
            luks_cipher: option_string_to_c_char(u.luks_cipher),
            luks_cipher_mode: option_string_to_c_char(u.luks_cipher_mode),
            luks_volume_key_size: u.luks_volume_key_size.unwrap_or(u64::MAX),
            luks_pbkdf_hash_algorithm: option_string_to_c_char(u.luks_pbkdf_hash_algorithm),
            luks_pbkdf_type: option_string_to_c_char(u.luks_pbkdf_type),
            luks_pbkdf_force_iterations: u.luks_pbkdf_force_iterations.unwrap_or(u64::MAX),
            luks_pbkdf_time_cost_usec: u.luks_pbkdf_time_cost_usec.unwrap_or(u64::MAX),
            luks_pbkdf_memory_cost: u.luks_pbkdf_memory_cost.unwrap_or(u64::MAX),
            luks_pbkdf_parallel_threads: u.luks_pbkdf_parallel_threads.unwrap_or(u64::MAX),
            luks_sector_size: u.luks_sector_size.unwrap_or(u64::MAX),
            auto_resize_mode: u.auto_resize_mode.into(),
            rebalance_weight: u.rebalance_weight.unwrap_or(u64::MAX),
            service: option_string_to_c_char(u.service),
            rate_limit_interval_usec: u.rate_limit_interval_usec.unwrap_or(0),
            rate_limit_interval_burst: u.rate_limit_interval_burst.unwrap_or(0),
            enforce_password_policy: u.enforce_password_policy,
            auto_login: u.auto_login,
            preferred_session_type: option_string_to_c_char(u.preferred_session_type),
            preferred_session_launcher: option_string_to_c_char(u.preferred_session_launcher),
            stop_delay_usec: u.stop_delay_usec.unwrap_or(u64::MAX),
            kill_processes: u.kill_processes,
            password_change_min_usec: u.password_change_min_usec.unwrap_or(u64::MAX),
            password_change_max_usec: u.password_change_max_usec.unwrap_or(u64::MAX),
            password_change_warn_usec: u.password_change_warn_usec.unwrap_or(u64::MAX),
            password_change_inactive_usec: u.password_change_inactive_usec.unwrap_or(u64::MAX),
            password_change_now: u.password_change_now,
            pkcs11_token_uri: u.pkcs11_token_uri.into(),
            fido2_hmac_credential: u.fido2_hmac_credential.into(),
            recovery_key_type: u.recovery_key_type.into(),
            self_modifiable_fields: u.self_modifiable_fields.into(),
            self_modifiable_blobs: u.self_modifiable_blobs.into(),
            self_modifiable_privileged: u.self_modifiable_privileged.into(),

            // Privileged section
            privileged: {
                match u.privileged {
                    Some(p) => {
                        let new: KirmesUserPrivileged = p.into();
                        Box::into_raw(Box::new(new))
                    },
                    None => ptr::null(),
                }
            },

            // PerMachine section
            per_machine: u.per_machine.into(),

            // Binding section
            binding: u.binding.into(),

            // Status section
            status: u.status.into(),

            // Signature section
            signature: u.signature.into(),

            // Secret section
            secret: {
                match u.secret {
                    Some(s) => {
                        let new: KirmesUserSecret = s.into();
                        Box::into_raw(Box::new(new))
                    },
                    None => ptr::null(),
                }
            },
        }
    }
}

impl Drop for KirmesUserRecord {
    fn drop(&mut self) {
        c_string_free!(self.user_name);
        c_string_free!(self.realm);
        c_string_free!(self.blob_directory);
        kirmes_boxed_vec_free!(self.blob_manifest);
        c_string_free!(self.real_name);
        c_string_free!(self.email_address);
        c_string_free!(self.icon_name);
        c_string_free!(self.location);
        c_string_free!(self.shell);
        kirmes_boxed_vec_free!(self.environment);
        c_string_free!(self.time_zone);
        c_string_free!(self.preferred_language);
        kirmes_str_vec_free!(self.additional_languages);
        c_string_free!(self.skeleton_directory);
        c_string_free!(self.cifs_domain);
        c_string_free!(self.cifs_user_name);
        c_string_free!(self.cifs_service);
        c_string_free!(self.cifs_extra_mount_options);
        c_string_free!(self.image_path);
        c_string_free!(self.home_directory);
        kirmes_str_vec_free!(self.member_of);
        c_string_free!(self.file_system_type);
        c_string_free!(self.partition_uuid);
        c_string_free!(self.luks_uuid);
        c_string_free!(self.file_system_uuid);
        c_string_free!(self.luks_extra_mount_options);
        c_string_free!(self.luks_cipher);
        c_string_free!(self.luks_cipher_mode);
        c_string_free!(self.luks_pbkdf_hash_algorithm);
        c_string_free!(self.luks_pbkdf_type);
        c_string_free!(self.service);
        c_string_free!(self.preferred_session_type);
        c_string_free!(self.preferred_session_launcher);
        kirmes_str_vec_free!(self.pkcs11_token_uri);
        kirmes_str_vec_free!(self.fido2_hmac_credential);
        kirmes_str_vec_free!(self.recovery_key_type);
        kirmes_str_vec_free!(self.self_modifiable_fields);
        kirmes_str_vec_free!(self.self_modifiable_blobs);
        kirmes_str_vec_free!(self.self_modifiable_privileged);
        box_free!(self.privileged);
        kirmes_boxed_vec_free!(self.per_machine);
        kirmes_boxed_vec_free!(self.binding);
        kirmes_boxed_vec_free!(self.status);
        kirmes_boxed_vec_free!(self.signature);
        box_free!(self.secret);
    }
}

/// Disposition of the user, i.e. the context it is defined in.
#[repr(C)]
#[derive(Debug)]
pub enum KirmesUserDisposition {
    /// root and nobody
    UserIntrinsic,
    /// statically allocated users for system services
    UserSystem,
    /// dynamically allocated users for system services
    UserDynamic,
    /// regular (typically human users)
    UserRegular,
    /// UID ranges allocated for container uses
    UserContainer,
    /// UID range allocated for foreign OS images
    UserForeign,
    /// Range above 2^31
    UserReserved,
    _UserDispositionMax,
}

impl From<UserDisposition> for KirmesUserDisposition {
    fn from(u: UserDisposition) -> Self {
        match u {
            UserDisposition::UserIntrinsic => KirmesUserDisposition::UserIntrinsic,
            UserDisposition::UserSystem => KirmesUserDisposition::UserSystem,
            UserDisposition::UserDynamic => KirmesUserDisposition::UserDynamic,
            UserDisposition::UserRegular => KirmesUserDisposition::UserRegular,
            UserDisposition::UserContainer => KirmesUserDisposition::UserContainer,
            UserDisposition::UserForeign => KirmesUserDisposition::UserForeign,
            UserDisposition::UserReserved => KirmesUserDisposition::UserReserved,
            UserDisposition::_UserDispositionMax => KirmesUserDisposition::_UserDispositionMax,
        }
    }
}

/// Indicates the storage mechanism for the user’s home directory.
#[repr(C)]
#[derive(Debug)]
pub enum KirmesUserHomeStorage {
    UserClassic,
    UserLuks,
    UserDirectory,
    UserSubvolume,
    UserFscrypt,
    UserCifs,
    _UserStorageMax,
}

impl From<UserHomeStorage> for KirmesUserHomeStorage {
    fn from(s: UserHomeStorage) -> Self {
        match s {
            UserHomeStorage::UserClassic => KirmesUserHomeStorage::UserClassic,
            UserHomeStorage::UserLuks => KirmesUserHomeStorage::UserLuks,
            UserHomeStorage::UserDirectory => KirmesUserHomeStorage::UserDirectory,
            UserHomeStorage::UserSubvolume => KirmesUserHomeStorage::UserSubvolume,
            UserHomeStorage::UserFscrypt => KirmesUserHomeStorage::UserFscrypt,
            UserHomeStorage::UserCifs => KirmesUserHomeStorage::UserCifs,
            UserHomeStorage::_UserStorageMax => KirmesUserHomeStorage::_UserStorageMax,
        }
    }
}

/// Controls whether the home area shall be grown automatically
#[repr(C)]
#[derive(Debug)]
pub enum KirmesAutoResizeMode {
    AutoResizeOff,
    AutoResizeGrow,
    AutoResizeShrinkAndGrow,
    _AutoResizeModeMax,
}

impl From<AutoResizeMode> for KirmesAutoResizeMode {
    fn from(m: AutoResizeMode) -> Self {
        match m {
            AutoResizeMode::AutoResizeOff => KirmesAutoResizeMode::AutoResizeOff,
            AutoResizeMode::AutoResizeGrow => KirmesAutoResizeMode::AutoResizeGrow,
            AutoResizeMode::AutoResizeShrinkAndGrow => {
                KirmesAutoResizeMode::AutoResizeShrinkAndGrow
            },
            AutoResizeMode::_AutoResizeModeMax => KirmesAutoResizeMode::_AutoResizeModeMax,
        }
    }
}

/// The `EncryptedKey` for PKCS#11
#[repr(C)]
#[derive(Debug)]
pub struct KirmesPkcs11EncryptedKey {
    /// PKCS#11 security token URI,
    pub uri: *const c_char,
    /// UNIX password hash to test the key against.
    pub hashed_password: *const c_char,
    /// Base64-encoded encrypted key
    pub data: *const c_char,
}

impl From<Pkcs11EncryptedKey> for KirmesPkcs11EncryptedKey {
    fn from(v: Pkcs11EncryptedKey) -> Self {
        KirmesPkcs11EncryptedKey {
            uri: string_to_c_char(v.uri),
            hashed_password: string_to_c_char(v.hashed_password),
            data: string_to_c_char(v.data),
        }
    }
}

impl Drop for KirmesPkcs11EncryptedKey {
    fn drop(&mut self) {
        c_string_free!(self.uri);
        c_string_free!(self.hashed_password);
        c_string_free!(self.data);
    }
}

#[repr(C)]
#[derive(Debug)]
pub struct KirmesPkcs11EncryptedKeyVec {
    pub vec: *const KirmesPkcs11EncryptedKey,
    pub size: usize,
}

impl From<Option<Vec<Pkcs11EncryptedKey>>> for KirmesPkcs11EncryptedKeyVec {
    fn from(v: Option<Vec<Pkcs11EncryptedKey>>) -> Self {
        match v {
            Some(vec) => {
                let mut out = vec
                    .into_iter()
                    .map(std::convert::Into::<KirmesPkcs11EncryptedKey>::into)
                    .collect::<Vec<_>>();
                out.shrink_to_fit();
                KirmesPkcs11EncryptedKeyVec {
                    size: out.len(),
                    vec: Box::into_raw(out.into_boxed_slice()).cast(),
                }
            },
            None => KirmesPkcs11EncryptedKeyVec {
                size: 0,
                vec: ptr::null(),
            },
        }
    }
}

/// Authentication support with FIDO2 devices that implement the hmac-secret extension
#[repr(C)]
#[derive(Debug)]
pub struct KirmesFido2HmacSalt {
    pub credential: *const c_char,
    pub salt: *const c_char,
    pub hashed_password: *const c_char,
    pub uv: bool,
    pub up: bool,
    pub client_pin: bool,
}

impl From<Fido2HmacSalt> for KirmesFido2HmacSalt {
    fn from(v: Fido2HmacSalt) -> Self {
        KirmesFido2HmacSalt {
            credential: string_to_c_char(v.credential),
            salt: string_to_c_char(v.salt),
            hashed_password: string_to_c_char(v.hashed_password),
            uv: v.uv,
            up: v.up,
            client_pin: v.client_pin,
        }
    }
}

impl Drop for KirmesFido2HmacSalt {
    fn drop(&mut self) {
        c_string_free!(self.credential);
        c_string_free!(self.salt);
        c_string_free!(self.hashed_password);
    }
}

#[repr(C)]
#[derive(Debug)]
pub struct KirmesFido2HmacSaltVec {
    pub vec: *const KirmesFido2HmacSalt,
    pub size: usize,
}

impl From<Option<Vec<Fido2HmacSalt>>> for KirmesFido2HmacSaltVec {
    fn from(v: Option<Vec<Fido2HmacSalt>>) -> Self {
        match v {
            Some(vec) => {
                let mut out = vec
                    .into_iter()
                    .map(std::convert::Into::<KirmesFido2HmacSalt>::into)
                    .collect::<Vec<_>>();
                out.shrink_to_fit();
                KirmesFido2HmacSaltVec {
                    size: out.len(),
                    vec: Box::into_raw(out.into_boxed_slice()).cast(),
                }
            },
            None => KirmesFido2HmacSaltVec {
                size: 0,
                vec: ptr::null(),
            },
        }
    }
}

/// Recovery keys for the user account
#[repr(C)]
#[derive(Debug)]
pub struct KirmesRecoveryKey {
    /// The type of recovery key, must be "modhex64" right now
    pub key_type: *const c_char,
    /// A UNIX password hash of the normalized form of modhex64
    pub hashed_password: *const c_char,
}

impl From<RecoveryKey> for KirmesRecoveryKey {
    fn from(key: RecoveryKey) -> Self {
        KirmesRecoveryKey {
            key_type: string_to_c_char(key.key_type),
            hashed_password: string_to_c_char(key.hashed_password),
        }
    }
}

impl Drop for KirmesRecoveryKey {
    fn drop(&mut self) {
        c_string_free!(self.key_type);
        c_string_free!(self.hashed_password);
    }
}

#[repr(C)]
#[derive(Debug)]
pub struct KirmesRecoveryKeyVec {
    pub vec: *const KirmesRecoveryKey,
    pub size: usize,
}

impl From<Option<Vec<RecoveryKey>>> for KirmesRecoveryKeyVec {
    fn from(v: Option<Vec<RecoveryKey>>) -> Self {
        match v {
            Some(vec) => {
                let mut out = vec
                    .into_iter()
                    .map(std::convert::Into::<KirmesRecoveryKey>::into)
                    .collect::<Vec<_>>();
                out.shrink_to_fit();
                KirmesRecoveryKeyVec {
                    size: out.len(),
                    vec: Box::into_raw(out.into_boxed_slice()).cast(),
                }
            },
            None => KirmesRecoveryKeyVec {
                size: 0,
                vec: ptr::null(),
            },
        }
    }
}

/// Contains information that the user and the administrator should be able to see, but other users
/// should not.
#[repr(C)]
#[derive(Debug)]
pub struct KirmesUserPrivileged {
    /// A user-selected password hint in free-form text.
    pub password_hint: *const c_char,
    /// An array of strings, each containing a hashed UNIX password string, in the format crypt(3)
    /// generates.
    pub hashed_passwords: KirmesStrVec,
    /// An array of strings, each listing an SSH public key that is authorized to access the
    /// account.
    pub ssh_authorized_keys: KirmesStrVec,
    /// An array of `Pkcs11EncryptedKey` objects.
    pub pkcs11_encrypted_key: KirmesPkcs11EncryptedKeyVec,
    /// An array of objects, implementing authentication support with FIDO2 devices that implement
    /// the hmac-secret extension.
    pub fido2_hmac_salt: KirmesFido2HmacSaltVec,
    /// An array of objects, each defining a recovery key.
    pub recovery_key: KirmesRecoveryKeyVec,
}

impl From<UserPrivileged> for KirmesUserPrivileged {
    fn from(v: UserPrivileged) -> Self {
        KirmesUserPrivileged {
            password_hint: option_string_to_c_char(v.password_hint),
            hashed_passwords: v.hashed_passwords.into(),
            ssh_authorized_keys: v.ssh_authorized_keys.into(),
            pkcs11_encrypted_key: v.pkcs11_encrypted_key.into(),
            fido2_hmac_salt: v.fido2_hmac_salt.into(),
            recovery_key: v.recovery_key.into(),
        }
    }
}

impl Drop for KirmesUserPrivileged {
    fn drop(&mut self) {
        c_string_free!(self.password_hint);
        kirmes_str_vec_free!(self.hashed_passwords);
        kirmes_str_vec_free!(self.ssh_authorized_keys);
        box_free!(self.pkcs11_encrypted_key.vec);
        box_free!(self.fido2_hmac_salt.vec);
        box_free!(self.recovery_key.vec);
    }
}

/// Resource management-related fields
#[repr(C)]
#[derive(Debug)]
pub struct KirmesUserPerMachine {
    /// An array of strings that are formatted 128-bit IDs in hex.
    pub match_machine_id: KirmesStrVec,
    /// An array of strings that are valid hostnames.
    pub match_hostname: KirmesStrVec,

    /// The absolute path to a world-readable copy of the user’s blob directory.
    pub blob_directory: *const c_char,
    /// An object, which maps valid blob directory filenames to SHA256 hashes formatted as hex
    /// strings.
    pub blob_manifest: KirmesStrVecMap,
    /// The name of an icon picked by the user.
    pub icon_name: *const c_char,
    /// A free-form location string describing the location of the user.
    pub location: *const c_char,
    /// The shell binary to use for terminal logins of this user.
    pub shell: *const c_char,
    /// The `umask` to set for the user’s login sessions.
    pub umask: u32,
    /// An array of strings, each containing an environment variable and its value to set for the
    /// user’s login session.
    pub environment: KirmesStrVecMap,
    /// The preferred timezone of the user.
    pub time_zone: *const c_char,
    /// The preferred language/locale for the user.
    pub preferred_language: *const c_char,
    /// The preferred languages/locales that should be used in the event that translations for the
    /// `preferredLanguage` are missing
    pub additional_languages: KirmesStrVec,
    /// An integer value in the range -20…19
    pub nice_level: i32,
    // An object, where each key refers to a Linux resource limit (such as RLIMIT_NOFILE and
    // similar).
    // TODO pub resource_limits: ???,
    /// If true, the user account is locked, the user may not log in.
    pub locked: bool,
    /// A timestamp when record should be considered invalid for the purpose of logging in (µs
    /// since the UNIX epoch 1970).
    pub not_before_usec: u64,
    /// The point in time after which logins shall not be permitted anymore (µs since the
    /// UNIX epoch 1970).
    pub not_after_usec: u64,
    /// The storage mechanism for the user’s home directory.
    pub storage: KirmesUserHomeStorage,
    /// The intended home directory disk space in bytes to assign to the user.
    pub disk_size: u64,
    /// Similar to diskSize but takes a relative value, but specifies a fraction of the available
    /// disk space on the selected storage medium to assign to the user.
    pub disk_size_relative: u64,
    /// The absolute path to the skeleton directory to populate a new home directory from.
    pub skeleton_directory: *const c_char,
    /// The UNIX access mask in the range 0…511 for the home directory when it is first created.
    pub access_mask: u32,
    /// The maximum number of tasks the user may start in parallel during system runtime.
    pub tasks_max: u64,
    /// Upper memory limits for all processes of the user.
    pub memory_high: u64,
    /// Upper memory limits for all processes of the user.
    pub memory_max: u64,
    /// CPU scheduling weights for the user’s processes (range 1…10000 ,defaults to
    /// 100).
    pub cpu_weight: u32,
    /// IO scheduling weights for the user’s processes (range 1…10000 ,defaults to 100).
    pub io_weight: u32,
    /// Control nosuid mount flag of the user’s home directories.
    pub mount_no_suid: bool,
    /// Control nodev mount flag of the user’s home directories.
    pub mount_no_dev: bool,
    /// Control noexec mount flag of the user’s home directories.
    pub mount_no_exec: bool,
    /// The Windows File Sharing domain (CIFS) to use.
    pub cifs_domain: *const c_char,
    /// The Windows File Sharing user name (CIFS) to associate this user record with.
    pub cifs_user_name: *const c_char,
    /// The Windows File Share service (CIFS) to mount as home directory of the user on login.
    pub cifs_service: *const c_char,
    /// Additional mount options to pass to mount.cifs when mounting the home directory CIFS share.
    pub cifs_extra_mount_options: *const c_char,
    /// An absolute file system path to the file, directory or block device to use for storage
    /// backing the home directory.
    pub image_path: *const c_char,
    /// The numeric UNIX user ID (UID) to use for the user.
    pub uid: u32,
    /// The numeric UNIX group ID (GID) to use for the user.
    pub gid: u32,
    /// An array of strings, each indicating a UNIX group this user shall be a member of.
    pub member_of: KirmesStrVec,
    /// One of ext4, xfs, btrfs (possibly others) to use as file system for the user’s home
    /// directory.
    pub file_system_type: *const c_char,
    /// A lower-case, text-formatted UUID, referencing the GPT partition UUID the home directory is
    /// located in.
    pub partition_uuid: *const c_char,
    /// A lower-case, text-formatted UUID, referencing the LUKS volume UUID the home directory is
    /// located in.
    pub luks_uuid: *const c_char,
    /// A lower-case, text-formatted UUID, referencing the file system UUID the home directory is
    /// located in.
    pub file_system_uuid: *const c_char,
    ///  If true and luks storage is used, controls whether the loopback block devices, LUKS and
    ///  the file system on top shall be used in discard mode.
    pub luks_discard: bool,
    /// Similar to luksDiscard, it controls whether to trim/allocate the file system/backing file
    /// when deactivating the home directory.
    pub luks_offline_discard: bool,
    /// The cipher to use for the LUKS storage mechanism.
    pub luks_cipher: *const c_char,
    // Selects the cipher mode to use for the LUKS storage mechanism.
    pub luks_cipher_mode: *const c_char,
    /// The volume key length in bytes to use for the LUKS storage mechanism.
    pub luks_volume_key_size: u64,
    /// Selects the hash algorithm to use for the PBKDF operation.
    pub luks_pbkdf_hash_algorithm: *const c_char,
    /// The PBKDF type to use for the LUKS storage mechanism.
    pub luks_pbkdf_type: *const c_char,
    /// The intended number of iterations for the PBKDF operation.
    pub luks_pbkdf_force_iterations: u64,
    /// The intended time cost for the PBKDF operation, in µs.
    pub luks_pbkdf_time_cost_usec: u64,
    /// The intended memory cost for the PBKDF operation, in bytes.
    pub luks_pbkdf_memory_cost: u64,
    /// The intended required parallel threads for the PBKDF operation.
    pub luks_pbkdf_parallel_threads: u64,
    /// The sector size to use for the LUKS storage mechanism, in bytes.
    pub luks_sector_size: u64,
    /// Controls whether the home area shall be grown automatically.
    pub auto_resize_mode: KirmesAutoResizeMode,
    /// Configures the free disk space rebalancing weight for the home area.
    pub rebalance_weight: u64,
    /// Configures the authentication rate limiting enforced on the user account.
    pub rate_limit_interval_usec: u64,
    /// Closely related to rateLimitIntervalUSec, that puts a limit on authentication attempts
    /// within the configured time interval.
    pub rate_limit_interval_burst: u64,
    /// Configures whether to enforce the system’s password policy when creating the home directory
    /// for the user or changing the user’s password.
    pub enforce_password_policy: bool,
    /// If true the user record is marked as suitable for auto-login.
    pub auto_login: bool,
    /// The user’s preferred session type (i.e. x11, wayland, or other values valid for
    /// `$XDG_SESSION_TYPE`).
    pub preferred_session_type: *const c_char,
    /// The user’s preferred session launcher desktop entry file (i.e. gnome, plasma, ...)
    pub preferred_session_launcher: *const c_char,
    /// The time in µs the per-user service manager is kept around after the user fully logged out.
    pub stop_delay_usec: u64,
    ///  If true all processes of the user are automatically killed when the user logs out.
    pub kill_processes: bool,
    /// Encoding how much time has to pass at least between password changes of the user.
    pub password_change_min_usec: u64,
    /// Encodes how much time has to pass at at most between password changes of the user.
    pub password_change_max_usec: u64,
    /// How much time to warn the user before their password expires, in µs.
    pub password_change_warn_usec: u64,
    /// Encodes how much time has to pass after the password expired that the account is
    /// deactivated, in µs.
    pub password_change_inactive_usec: u64,
    /// If true the user has to change their password on next login.
    pub password_change_now: bool,
    /// An array of strings, each with an RFC 7512 compliant PKCS#11 URI referring to security
    /// token (or smart card) of some form, that shall be associated with the user and may be used
    /// for authentication.
    pub pkcs11_token_uri: KirmesStrVec,
    /// An array of strings,each with a Base64-encoded FIDO2 credential ID that shall be used for
    /// authentication with FIDO2 devices that implement the hmac-secret extension.
    pub fido2_hmac_credential: KirmesStrVec,
    /// An array of strings, each corresponding to a field name that can appear in the regular or
    /// perMachine sections.
    pub self_modifiable_fields: KirmesStrVec,
    /// Similar to selfModifiableFields, but it lists blobs that the user is allowed to edit.
    pub self_modifiable_blobs: KirmesStrVec,
    /// Similar to selfModifiableFields, but it lists fields in the privileged section that the
    /// user is allowed to edit.
    pub self_modifiable_privileged: KirmesStrVec,
}

impl From<UserPerMachine> for KirmesUserPerMachine {
    fn from(m: UserPerMachine) -> Self {
        KirmesUserPerMachine {
            match_machine_id: m.match_machine_id.into(),
            match_hostname: m.match_hostname.into(),

            blob_directory: option_string_to_c_char(m.blob_directory),
            blob_manifest: m.blob_manifest.into(),
            icon_name: option_string_to_c_char(m.icon_name),
            location: option_string_to_c_char(m.location),
            shell: option_string_to_c_char(m.shell),
            umask: m.umask.unwrap_or(u32::MAX),
            environment: m.environment.into(),
            time_zone: option_string_to_c_char(m.time_zone),
            preferred_language: option_string_to_c_char(m.preferred_language),
            additional_languages: m.additional_languages.into(),
            nice_level: m.nice_level.unwrap_or(i32::MAX),
            locked: m.locked,
            not_before_usec: m.not_before_usec.unwrap_or(u64::MAX),
            not_after_usec: m.not_after_usec.unwrap_or(u64::MAX),
            storage: m.storage.into(),
            disk_size: m.disk_size.unwrap_or(u64::MAX),
            disk_size_relative: m.disk_size_relative.unwrap_or(u64::MAX),
            skeleton_directory: option_string_to_c_char(m.skeleton_directory),
            access_mask: m.access_mask.unwrap_or(u32::MAX),
            tasks_max: m.tasks_max.unwrap_or(u64::MAX),
            memory_high: m.memory_high.unwrap_or(u64::MAX),
            memory_max: m.memory_max.unwrap_or(u64::MAX),
            cpu_weight: m.cpu_weight.unwrap_or(u32::MAX),
            io_weight: m.io_weight.unwrap_or(u32::MAX),
            mount_no_suid: m.mount_no_suid,
            mount_no_dev: m.mount_no_dev,
            mount_no_exec: m.mount_no_exec,
            cifs_domain: option_string_to_c_char(m.cifs_domain),
            cifs_user_name: option_string_to_c_char(m.cifs_user_name),
            cifs_service: option_string_to_c_char(m.cifs_service),
            cifs_extra_mount_options: option_string_to_c_char(m.cifs_extra_mount_options),
            image_path: option_string_to_c_char(m.image_path),
            uid: m.uid.unwrap_or(u32::MAX),
            gid: m.gid.unwrap_or(u32::MAX),
            member_of: m.member_of.into(),
            file_system_type: option_string_to_c_char(m.file_system_type),
            partition_uuid: option_string_to_c_char(m.partition_uuid),
            luks_uuid: option_string_to_c_char(m.luks_uuid),
            file_system_uuid: option_string_to_c_char(m.file_system_uuid),
            luks_discard: m.luks_discard,
            luks_offline_discard: m.luks_offline_discard,
            luks_cipher: option_string_to_c_char(m.luks_cipher),
            luks_cipher_mode: option_string_to_c_char(m.luks_cipher_mode),
            luks_volume_key_size: m.luks_volume_key_size.unwrap_or(u64::MAX),
            luks_pbkdf_hash_algorithm: option_string_to_c_char(m.luks_pbkdf_hash_algorithm),
            luks_pbkdf_type: option_string_to_c_char(m.luks_pbkdf_type),
            luks_pbkdf_force_iterations: m.luks_pbkdf_force_iterations.unwrap_or(u64::MAX),
            luks_pbkdf_time_cost_usec: m.luks_pbkdf_time_cost_usec.unwrap_or(u64::MAX),
            luks_pbkdf_memory_cost: m.luks_pbkdf_memory_cost.unwrap_or(u64::MAX),
            luks_pbkdf_parallel_threads: m.luks_pbkdf_parallel_threads.unwrap_or(u64::MAX),
            luks_sector_size: m.luks_sector_size.unwrap_or(u64::MAX),
            auto_resize_mode: m.auto_resize_mode.into(),
            rebalance_weight: m.rebalance_weight.unwrap_or(u64::MAX),
            rate_limit_interval_usec: m.rate_limit_interval_usec.unwrap_or(0),
            rate_limit_interval_burst: m.rate_limit_interval_burst.unwrap_or(0),
            enforce_password_policy: m.enforce_password_policy,
            auto_login: m.auto_login,
            preferred_session_type: option_string_to_c_char(m.preferred_session_type),
            preferred_session_launcher: option_string_to_c_char(m.preferred_session_launcher),
            stop_delay_usec: m.stop_delay_usec.unwrap_or(u64::MAX),
            kill_processes: m.kill_processes,
            password_change_min_usec: m.password_change_min_usec.unwrap_or(u64::MAX),
            password_change_max_usec: m.password_change_max_usec.unwrap_or(u64::MAX),
            password_change_warn_usec: m.password_change_warn_usec.unwrap_or(u64::MAX),
            password_change_inactive_usec: m.password_change_inactive_usec.unwrap_or(u64::MAX),
            password_change_now: m.password_change_now,
            pkcs11_token_uri: m.pkcs11_token_uri.into(),
            fido2_hmac_credential: m.fido2_hmac_credential.into(),
            self_modifiable_fields: m.self_modifiable_fields.into(),
            self_modifiable_blobs: m.self_modifiable_blobs.into(),
            self_modifiable_privileged: m.self_modifiable_privileged.into(),
        }
    }
}

/// A structure holding an array of `struct KirmesUserPerMachine`
///
/// Examples
///
/// ```custom,{class=language-c}
/// KirmesUserPerMachineVec v;
/// size_t i;
/// const char *my_machine_id = get_machine_id();
/// const char *my_hostname = get_hostname();
///
/// for (i = 0; i < v.size; i++) {
///     struct KirmesUserPerMachine *per_machine = &value.vec[i];
///     bool for_me = false;
///     size_t j;
///
///     for (j = 0; j < per_machine->match_machine_id.size; j++) {
///         const char *id = per_machine->match_machine_id.vec[j];
///         int cmp;
///
///         cmp = strcasecmp(id, my_machine_id);
///         if (cmp == 0) {
///             for_me = true;
///         }
///     }
///
///     for (j = 0; j < per_machine->match_hostname.size; j++) {
///         const char *host = per_machine->match_hostname.vec[j];
///         int cmp;
///
///         cmp = strcasecmp(host, my_hostname);
///         if (cmp == 0) {
///             for_me = true;
///         }
///     }
///
///     if (!for_me) {
///         continue;
///     }
///
///     // This per_machine entry is for this host
/// }
/// ```
///
#[repr(C)]
#[derive(Debug)]
pub struct KirmesUserPerMachineVec {
    pub vec: *const KirmesUserPerMachine,
    pub size: usize,
}

impl From<Option<Vec<UserPerMachine>>> for KirmesUserPerMachineVec {
    fn from(value: Option<Vec<UserPerMachine>>) -> Self {
        match value {
            Some(val) => {
                let mut out: Vec<KirmesUserPerMachine> = val
                    .into_iter()
                    .map(std::convert::Into::<KirmesUserPerMachine>::into)
                    .collect();
                out.shrink_to_fit();

                KirmesUserPerMachineVec {
                    size: out.len(),
                    vec: Box::into_raw(out.into_boxed_slice()).cast(),
                }
            },
            None => KirmesUserPerMachineVec {
                size: 0,
                vec: ptr::null(),
            },
        }
    }
}

/// The binding section overrides their equally named counterparts in the regular and perMachine
/// sections.
#[repr(C)]
#[derive(Debug)]
pub struct KirmesUserBinding {
    /// The machine ID the binding is intended for (key).
    pub for_machine_id: *const c_char,

    /// The absolute path to a world-readable copy of the user’s blob directory.
    pub blob_directory: *const c_char,
    /// An absolute file system path to the file, directory or block device to use for storage
    /// backing the home directory.
    pub image_path: *const c_char,
    /// An absolute file system path to the home directory.
    pub home_directory: *const c_char,
    /// A lower-case, text-formatted UUID, referencing the GPT partition UUID the home directory is
    /// located in.
    pub partition_uuid: *const c_char,
    /// A lower-case, text-formatted UUID, referencing the LUKS volume UUID the home directory is
    /// located in.
    pub luks_uuid: *const c_char,
    /// One of ext4, xfs, btrfs (possibly others) to use as file system for the user’s home
    /// directory.
    pub file_system_type: *const c_char,
    /// A lower-case, text-formatted UUID, referencing the GPT partition UUID the home directory is
    /// located in.
    pub file_system_uuid: *const c_char,
    /// The numeric UNIX user ID (UID) to use for the user.
    pub uid: u32,
    /// The numeric UNIX group ID (GID) to use for the user.
    pub gid: u32,
    /// The storage mechanism for the user’s home directory.
    pub storage: KirmesUserHomeStorage,
    /// The cipher to use for the LUKS storage mechanism.
    pub luks_cipher: *const c_char,
    // Selects the cipher mode to use for the LUKS storage mechanism.
    pub luks_cipher_mode: *const c_char,
    /// The volume key length in bytes to use for the LUKS storage mechanism.
    pub luks_volume_key_size: u64,
}

impl From<UserBinding> for KirmesUserBinding {
    fn from(b: UserBinding) -> Self {
        KirmesUserBinding {
            for_machine_id: ptr::null(),

            blob_directory: option_string_to_c_char(b.blob_directory),
            image_path: option_string_to_c_char(b.image_path),
            home_directory: option_string_to_c_char(b.home_directory),
            partition_uuid: option_string_to_c_char(b.partition_uuid),
            luks_uuid: option_string_to_c_char(b.luks_uuid),
            file_system_type: option_string_to_c_char(b.file_system_type),
            file_system_uuid: option_string_to_c_char(b.file_system_uuid),
            uid: b.uid.unwrap_or(u32::MAX),
            gid: b.gid.unwrap_or(u32::MAX),
            storage: b.storage.into(),
            luks_cipher: option_string_to_c_char(b.luks_cipher),
            luks_cipher_mode: option_string_to_c_char(b.luks_cipher_mode),
            luks_volume_key_size: b.luks_volume_key_size.unwrap_or(u64::MAX),
        }
    }
}

impl Drop for KirmesUserBinding {
    fn drop(&mut self) {
        c_string_free!(self.for_machine_id);

        c_string_free!(self.blob_directory);
        c_string_free!(self.image_path);
        c_string_free!(self.home_directory);
        c_string_free!(self.partition_uuid);
        c_string_free!(self.luks_uuid);
        c_string_free!(self.file_system_type);
        c_string_free!(self.file_system_uuid);
        c_string_free!(self.luks_cipher);
        c_string_free!(self.luks_cipher_mode);
    }
}

/// Structure for an array of `struct KirmesUserBinding`.
///
/// # Examples
///
/// ```custom,{class=language-c}
/// KirmesUserBindingVec v;
/// size_t i;
/// const char *my_machine_id = get_machine_id();
///
/// for (i = 0; i < v.size; i++) {
///     struct KirmesUserBinding *binding = &value.vec[i];
///     int cmp;
///
///     // Check if this entry is for this machine
///     cmp = strcasecmp(binding->for_machine_id, my_machine_id);
///     if (cmp != 0) {
///         continue;
///     }
///
///     // This binding entry is for this machine
/// }
/// ```
#[repr(C)]
#[derive(Debug)]
pub struct KirmesUserBindingVec {
    pub vec: *const KirmesUserBinding,
    pub size: usize,
}

impl From<Option<HashMap<String, UserBinding>>> for KirmesUserBindingVec {
    fn from(value: Option<HashMap<String, UserBinding>>) -> Self {
        match value {
            Some(val) => {
                let mut out: Vec<KirmesUserBinding> = val
                    .into_iter()
                    .map(|(k, v)| {
                        let mut b: KirmesUserBinding = v.into();
                        b.for_machine_id = string_to_c_char(k);
                        b
                    })
                    .collect();
                out.shrink_to_fit();

                KirmesUserBindingVec {
                    size: out.len(),
                    vec: Box::into_raw(out.into_boxed_slice()).cast(),
                }
            },
            None => KirmesUserBindingVec {
                size: 0,
                vec: ptr::null(),
            },
        }
    }
}

/// Contains information about current runtime resource usage.
#[repr(C)]
#[derive(Debug)]
pub struct KirmesUserStatus {
    /// The machine ID the binding is intended for (key).
    pub for_machine_id: *const c_char,

    /// The currently used disk space of the home directory in bytes.
    pub disk_usage: u64,
    /// The number of 'free' bytes in the disk space allotment.
    pub disk_free: u64,
    /// The intended home directory disk space in bytes to assign to the user.
    pub disk_size: u64,
    /// The upper bounds when changing the `diskSize` value, in bytes.
    pub disk_ceiling: u64,
    /// The lower bounds when changing the `diskSize` value, in bytes.
    pub disk_floor: u64,
    /// The current state of the home directory.
    pub state: *const c_char,
    /// If true indicates that the user record is signed by a public key for which the private key
    /// is available locally.
    pub signed_locally: bool,
    /// This counter is increased by one on every successful authentication attempt.
    pub good_authentication_counter: u64,
    /// This counter is increased by one on every unsuccessfully authentication attempt.
    pub bad_authentication_counter: u64,
    /// The time of the last successful authentication attempt in µs since the UNIX epoch (1970).
    pub last_good_authentication_usec: u64,
    /// Similar, but the timestamp of the last unsuccessfully authentication attempt.
    pub last_bad_authentication_usec: u64,
    /// The µs timestamp since the UNIX epoch (1970) where the most recent rate limiting interval
    /// has been started.
    pub rate_limit_begin_usec: u64,
    /// Counting the authentication attempts in the current rate limiting interval.
    pub rate_limit_count: u64,
    /// If true the manager of this user record determined the home directory being on removable
    /// media.
    pub removeable: bool,
    /// The service that defines or manages this user record.
    pub service: *const c_char,
}

impl From<UserStatus> for KirmesUserStatus {
    fn from(s: UserStatus) -> Self {
        KirmesUserStatus {
            for_machine_id: ptr::null(),

            disk_usage: s.disk_usage.unwrap_or(u64::MAX),
            disk_free: s.disk_free.unwrap_or(u64::MAX),
            disk_size: s.disk_size.unwrap_or(u64::MAX),
            disk_ceiling: s.disk_ceiling.unwrap_or(u64::MAX),
            disk_floor: s.disk_floor.unwrap_or(u64::MAX),
            state: option_string_to_c_char(s.state),
            signed_locally: s.signed_locally,
            good_authentication_counter: s.good_authentication_counter.unwrap_or(u64::MAX),
            bad_authentication_counter: s.bad_authentication_counter.unwrap_or(u64::MAX),
            last_bad_authentication_usec: s.bad_authentication_counter.unwrap_or(u64::MAX),
            last_good_authentication_usec: s.good_authentication_counter.unwrap_or(u64::MAX),
            rate_limit_begin_usec: s.rate_limit_begin_usec.unwrap_or(u64::MAX),
            rate_limit_count: s.rate_limit_count.unwrap_or(u64::MAX),
            removeable: s.removeable,
            service: option_string_to_c_char(s.service),
        }
    }
}

impl Drop for KirmesUserStatus {
    fn drop(&mut self) {
        c_string_free!(self.for_machine_id);
        c_string_free!(self.state);
        c_string_free!(self.service);
    }
}

/// Structure tho hold an array of `struct KirmesUserStatus`.
///
/// # Examples
///
/// ```custom,{class=language-c}
/// KirmesUserStatusVec v;
/// size_t i;
/// const char *my_machine_id = get_machine_id();
///
/// for (i = 0; i < v.size; i++) {
///     struct KirmesUserStatus *status = &value.vec[i];
///     int cmp;
///
///     // Check if this entry is for this machine
///     cmp = strcasecmp(status->for_machine_id, my_machine_id);
///     if (cmp != 0) {
///         continue;
///     }
///
///     // This status entry is for this machine
/// }
/// ```
#[repr(C)]
#[derive(Debug)]
pub struct KirmesUserStatusVec {
    pub vec: *const KirmesUserStatus,
    pub size: usize,
}

impl From<Option<HashMap<String, UserStatus>>> for KirmesUserStatusVec {
    fn from(value: Option<HashMap<String, UserStatus>>) -> Self {
        match value {
            Some(val) => {
                let mut out: Vec<KirmesUserStatus> = val
                    .into_iter()
                    .map(|(k, v)| {
                        let mut s: KirmesUserStatus = v.into();
                        s.for_machine_id = string_to_c_char(k);
                        s
                    })
                    .collect();
                out.shrink_to_fit();

                KirmesUserStatusVec {
                    size: out.len(),
                    vec: Box::into_raw(out.into_boxed_slice()).cast(),
                }
            },
            None => KirmesUserStatusVec {
                size: 0,
                vec: ptr::null(),
            },
        }
    }
}

/// Contains one or more cryptographic signatures of a reduced version of the user record.
#[repr(C)]
#[derive(Debug)]
pub struct KirmesUserSignature {
    /// The signature, encoded in Base64
    pub data: *const c_char,
    /// A copy of the public key whose private key was used to make the signature, in PEM format.
    pub key: *const c_char,
}

impl From<UserSignature> for KirmesUserSignature {
    fn from(s: UserSignature) -> Self {
        KirmesUserSignature {
            data: string_to_c_char(s.data),
            key: string_to_c_char(s.key),
        }
    }
}

impl Drop for KirmesUserSignature {
    fn drop(&mut self) {
        c_string_free!(self.data);
        c_string_free!(self.key);
    }
}

/// Structure for an array of `struct KirmesUserSignature`
///
/// Examples
///
/// ```custom,{class=language-c}
/// KirmesUserSignatureVec v;
/// size_t i;
///
/// for (i = 0; i < v.size; i++) {
///     struct KirmesUserSignature *signature = &value.vec[i];
///
///     // Access signature data and key
/// }
///
#[repr(C)]
#[derive(Debug)]
pub struct KirmesUserSignatureVec {
    pub vec: *const KirmesUserSignature,
    pub size: usize,
}

impl From<Option<Vec<UserSignature>>> for KirmesUserSignatureVec {
    fn from(value: Option<Vec<UserSignature>>) -> Self {
        match value {
            Some(val) => {
                let mut out: Vec<KirmesUserSignature> = val
                    .into_iter()
                    .map(std::convert::Into::<KirmesUserSignature>::into)
                    .collect();
                out.shrink_to_fit();

                KirmesUserSignatureVec {
                    size: out.len(),
                    vec: Box::into_raw(out.into_boxed_slice()).cast(),
                }
            },
            None => KirmesUserSignatureVec {
                size: 0,
                vec: ptr::null(),
            },
        }
    }
}

/// Secret user credentials, such as password or PIN information.
#[repr(C)]
#[derive(Debug)]
pub struct KirmesUserSecret {
    /// An array of strings, each containing a plain text password.
    pub passwords: KirmesStrVec,
    /// An array of strings, each containing a plain text PIN, suitable for unlocking security
    /// tokens that require that.
    pub token_pin: KirmesStrVec,
    /// If set to true allows the receiver to use the PKCS#11 “protected authentication path” for
    /// authenticating the user.
    pub pkcs11_protected_authentication_path_permitted: bool,
    /// If set to true allows the receiver to use the FIDO2 “user presence” flag.
    pub fido2_user_precense_permitted: bool,
    /// If set to true allows the receiver to use the FIDO2 “user verification” flag.
    pub fido2_user_verification_permitted: bool,
}

impl From<UserSecret> for KirmesUserSecret {
    fn from(s: UserSecret) -> Self {
        KirmesUserSecret {
            passwords: s.passwords.into(),
            token_pin: s.token_pin.into(),
            pkcs11_protected_authentication_path_permitted: s
                .pkcs11_protected_authentication_path_permitted,
            fido2_user_precense_permitted: s.fido2_user_precense_permitted,
            fido2_user_verification_permitted: s.fido2_user_verification_permitted,
        }
    }
}

impl Drop for KirmesUserSecret {
    fn drop(&mut self) {
        kirmes_str_vec_free!(self.passwords);
        kirmes_str_vec_free!(self.token_pin);
    }
}
