/*
 * Copyright (c) 2025      Andreas Schneider <asn@samba.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

use std::collections::HashMap;
use std::ffi::{CString, c_char};
use std::ptr;

/// A structure holding a string array
///
/// # Examples
///
/// ```custom,{class=language-c}
/// KirmesStrVec v;
/// size_t i;
///
/// for (i = 0; i < v.size; i++) {
///     const char *x = value.vec[i];
///     printf("%s\n", x);
/// }
/// ```
#[repr(C)]
#[derive(Debug)]
pub struct KirmesStrVec {
    /// The string array
    pub vec: *const *const c_char,
    /// The number of elements in the string array
    pub size: usize,
}

impl From<Option<Vec<String>>> for KirmesStrVec {
    fn from(value: Option<Vec<String>>) -> Self {
        match value {
            Some(vec) => {
                let mut out = vec.into_iter().map(string_to_c_char).collect::<Vec<_>>();
                out.shrink_to_fit(); // Don't waste space
                KirmesStrVec {
                    size: out.len(),
                    vec: Box::into_raw(out.into_boxed_slice()).cast(),
                }
            },
            None => KirmesStrVec {
                vec: ptr::null(),
                size: 0,
            },
        }
    }
}

macro_rules! kirmes_str_vec_free {
    ($input:expr) => {{
        if $input.size > 0 {
            unsafe {
                let vec: Vec<*const c_char> = Box::from_raw(ptr::slice_from_raw_parts_mut(
                    $input.vec.cast_mut(),
                    $input.size,
                ))
                .into_vec();
                for v in vec {
                    let _ = CString::from_raw(v.cast_mut());
                }
            }
        }
    }};
}
pub(crate) use kirmes_str_vec_free;

/// Key/Value mapping for `HashMap<String, String>`
#[repr(C)]
#[derive(Debug)]
pub struct KirmesStrMap {
    pub key: *const c_char,
    pub val: *const c_char,
}

impl Drop for KirmesStrMap {
    fn drop(&mut self) {
        c_string_free!(self.key);
        c_string_free!(self.val);
    }
}

/// An array holding `struct KirmesStrMap`
/// # Examples
///
/// ```custom,{class=language-c}
/// KirmesStrVecMap m;
/// size_t i;
///
/// for (i = 0; i < m.size; i++) {
///     const char *key = value.vec[i].key;
///     const char *val = value.vec[i].val;
///     printf("%s: %s\n", key, val);
/// }
/// ```
#[repr(C)]
#[derive(Debug)]
pub struct KirmesStrVecMap {
    pub vec: *const KirmesStrMap,
    pub size: usize,
}

impl From<Option<HashMap<String, String>>> for KirmesStrVecMap {
    fn from(value: Option<HashMap<String, String>>) -> Self {
        match value {
            Some(val) => {
                let mut out: Vec<KirmesStrMap> = val
                    .into_iter()
                    .map(|(k, v)| KirmesStrMap {
                        key: string_to_c_char(k),
                        val: string_to_c_char(v),
                    })
                    .collect();
                out.shrink_to_fit();

                KirmesStrVecMap {
                    size: out.len(),
                    vec: Box::into_raw(out.into_boxed_slice()).cast(),
                }
            },
            None => KirmesStrVecMap {
                vec: ptr::null(),
                size: 0,
            },
        }
    }
}

macro_rules! kirmes_boxed_vec_free {
    ($input:expr) => {{
        if $input.size > 0 {
            unsafe {
                let _: Vec<_> = Box::from_raw(ptr::slice_from_raw_parts_mut(
                    $input.vec.cast_mut(),
                    $input.size,
                ))
                .into_vec();
            }
        }
    }};
}
pub(crate) use kirmes_boxed_vec_free;

pub fn string_to_c_char(s: String) -> *const c_char {
    match CString::new(s) {
        Ok(cstr) => cstr.into_raw(),
        Err(_) => ptr::null(),
    }
}

pub fn option_string_to_c_char(s: Option<String>) -> *const c_char {
    match s {
        Some(str) => string_to_c_char(str),
        None => ptr::null(),
    }
}

macro_rules! c_string_free {
    ($input:expr) => {{
        if !$input.is_null() {
            unsafe {
                let _ = CString::from_raw($input.cast_mut());
            }
        }
    }};
}
pub(crate) use c_string_free;
