/*
 * Copyright (c) 2025      Andreas Schneider <asn@samba.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */
use crate::io_systemd_UserDatabase;
use tracing::error;

pub type Result<T> = core::result::Result<T, Error>;

/// This type represents all possible errors that can occur when dealing with records
#[derive(Debug /*, From */)]
pub enum Error {
    NoRecordFound,
    BadService,
    ServiceNotAvailable,
    ConflictingRecordFound,
    EnumerationNotSupported,
    InvalidRecord,
    InternalError,
}

impl core::fmt::Display for Error {
    fn fmt(
        &self,
        fmt: &mut core::fmt::Formatter<'_>,
    ) -> core::result::Result<(), core::fmt::Error> {
        write!(fmt, "{self:?}")
    }
}

impl std::error::Error for Error {}

impl From<io_systemd_UserDatabase::Error> for Error {
    fn from(e: io_systemd_UserDatabase::Error) -> Self {
        match e.0 {
            io_systemd_UserDatabase::ErrorKind::Varlink_Error => {
                if e.1.is_some() {
                    error!("{:?}: {}", e.1, e.2.unwrap_or(""));
                }
                Error::InternalError
            },
            io_systemd_UserDatabase::ErrorKind::VarlinkReply_Error => {
                if e.1.is_some() {
                    error!("{:?}: {}", e.1, e.2.unwrap_or(""));
                }
                Error::InternalError
            },
            io_systemd_UserDatabase::ErrorKind::ConflictingRecordFound(v) => {
                if v.is_some() {
                    error!("{:?}", v);
                }
                Error::ConflictingRecordFound
            },
            io_systemd_UserDatabase::ErrorKind::BadService(v) => {
                if v.is_some() {
                    error!("{:?}", v);
                }
                Error::BadService
            },
            io_systemd_UserDatabase::ErrorKind::EnumerationNotSupported(v) => {
                if v.is_some() {
                    error!("{:?}", v);
                }
                Error::EnumerationNotSupported
            },
            io_systemd_UserDatabase::ErrorKind::NoRecordFound(v) => {
                if v.is_some() {
                    error!("{:?}", v);
                }
                Error::NoRecordFound
            },
            io_systemd_UserDatabase::ErrorKind::ServiceNotAvailable(v) => {
                if v.is_some() {
                    error!("{:?}", v);
                }
                Error::ServiceNotAvailable
            },
        }
    }
}
