/*
 * Copyright (c) 2025      Andreas Schneider <asn@samba.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

use derivative::Derivative;
use serde::Deserialize;
use serde_json::Value;
use std::collections::HashMap;
use tracing::error;

/// Disposition of the group, i.e. the context it is defined in.
#[derive(Derivative, Deserialize)]
#[derivative(Default(bound = ""), Debug)]
pub enum GroupDisposition {
    /// root and nobody
    #[serde(rename = "intrinsic")]
    GroupIntrinsic,
    /// statically allocated groups for system services
    #[serde(rename = "system")]
    GroupSystem,
    /// dynamically allocated groups for system services
    #[serde(rename = "dynamic")]
    GroupDynamic,
    /// regular (typically human groups)
    #[derivative(Default)]
    #[serde(rename = "regular")]
    GroupRegular,
    /// UID ranges allocated for container uses
    #[serde(rename = "container")]
    GroupContainer,
    /// UID range allocated for foreign OS images
    #[serde(rename = "foreign")]
    GroupForeign,
    /// Range above 2^31
    #[serde(rename = "reserved")]
    GroupReserved,
    _GroupDispositionMax,
}

#[derive(Deserialize, Derivative)]
#[derivative(Debug, Default)]
pub struct GroupPrivileged {
    /// An array of strings with UNIX hashed passwords.
    #[serde(default, rename = "hashedPassword")]
    pub hashed_passwords: Option<Vec<String>>,
}

/// Resource management-related fields
#[derive(Deserialize, Derivative)]
#[derivative(Debug, Default)]
pub struct GroupPerMachine {
    /// An array of strings that are formatted 128-bit IDs in hex.
    #[serde(default)]
    pub match_machine_id: Option<Vec<String>>,
    /// An array of strings that are valid hostnames.
    #[serde(default)]
    pub match_hostname: Option<Vec<String>>,

    /// The numeric UNIX group ID (GID) to use for the group.
    #[serde(default)]
    pub gid: Option<u32>,
    /// An array of strings, listing user names that are members of this group.
    #[serde(default)]
    pub members: Option<Vec<String>>,
    /// An array of strings, listing user names that shall be considered “administrators” of this
    /// group.
    #[serde(default)]
    pub administrators: Option<Vec<String>>,
}

#[derive(Deserialize, Derivative)]
#[derivative(Debug, Default)]
pub struct GroupBinding {
    /// The numeric UNIX group ID (GID) to use for the group.
    #[serde(default)]
    pub gid: Option<u32>,
}

#[derive(Deserialize, Derivative)]
#[derivative(Debug, Default)]
pub struct GroupStatus {
    /// The service that defines or manages this group record.
    #[serde(default)]
    pub service: Option<String>,
}

/// Contains one or more cryptographic signatures of a reduced version of the group record.
#[derive(Deserialize, Derivative)]
#[derivative(Debug, Default)]
pub struct GroupSignature {
    /// The signature, encoded in Base64
    #[serde(default)]
    pub data: String,
    /// A copy of the public key whose private key was used to make the signature, in PEM format.
    #[serde(default)]
    pub key: String,
}

/// Currently no fields are defined in this section for group records.
#[derive(Deserialize, Derivative)]
#[derivative(Debug, Default)]
pub struct GroupSecret {}

/// Group Records
///
/// For a detailed explanation please visit <https://systemd.io/GROUP_RECORDS>
#[allow(clippy::struct_excessive_bools)]
#[derive(Deserialize, Derivative)]
#[derivative(Debug, Default)]
pub struct GroupRecord {
    //
    // Regular section
    //
    /// The UNIX group name for this record.
    #[serde(rename = "groupName")]
    pub group_name: String,
    /// The 'realm' a group is defined in.
    #[serde(default)]
    pub realm: Option<String>,
    /// Clarifies the disposition of the group, i.e. the context it is defined in.
    #[derivative(Default(value = "GroupDisposition::GroupRegular"))]
    #[serde(default)]
    pub disposition: GroupDisposition,
    /// The service that defines or manages this group record.
    #[serde(default)]
    pub service: Option<String>,
    /// A timestamp indicating when the group record was last changed (µs since the epoch 1970).
    #[serde(default, rename = "lastChangeUSec")]
    pub last_change_usec: Option<u64>,
    /// The numeric UNIX group ID (GID) to use for the group.
    #[serde(default)]
    pub gid: Option<u32>,
    /// An array of strings, listing user names that are members of this group.
    #[serde(default)]
    pub members: Option<Vec<String>>,
    /// An array of strings, listing user names that shall be considered “administrators” of this
    /// group.
    #[serde(default)]
    pub administrators: Option<Vec<String>>,

    //
    // Privileged section
    //
    #[serde(default)]
    privileged: Option<GroupPrivileged>,

    //
    // PerMachine section
    //
    per_machine: Option<HashMap<String, GroupPerMachine>>,

    //
    // Binding section
    //
    #[serde(default)]
    binding: Option<HashMap<String, GroupBinding>>,

    //
    // Status section
    //
    #[serde(default)]
    status: Option<HashMap<String, GroupStatus>>,

    //
    // Signature section
    //
    #[serde(default)]
    pub signature: Option<Vec<GroupSignature>>,

    //
    // Secret section
    //
    pub secret: Option<GroupSecret>,
}

// Trait to convert from json Value to UserRecord
impl TryFrom<serde_json::Value> for GroupRecord {
    type Error = crate::Error;

    fn try_from(value: Value) -> Result<Self, Self::Error> {
        let group_record: GroupRecord = match serde_json::from_value(value) {
            Ok(g) => g,
            Err(e) => {
                error!("Failed to deserialize serde_json::Value: {e:?}");
                return Err(crate::Error::InvalidRecord);
            },
        };

        Ok(group_record)
    }
}

impl core::fmt::Display for GroupRecord {
    fn fmt(
        &self,
        fmt: &mut core::fmt::Formatter<'_>,
    ) -> core::result::Result<(), core::fmt::Error> {
        writeln!(fmt, "Group Name:  {}", &self.group_name)?;
        writeln!(fmt, "Disposition: {:?}", &self.disposition)?;
        if let Some(gid) = &self.gid {
            writeln!(fmt, "GID:         {gid}")?;
        }
        if let Some(members) = &self.members {
            write!(fmt, "Groups:      ")?;
            let mut first = true;
            for member in members {
                if first {
                    writeln!(fmt, "{member}")?;
                    first = false;
                } else {
                    writeln!(fmt, "             {member}")?;
                }
            }
        }

        Ok(())
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use serde_json::json;
    //use tracing::Level;
    //use tracing_subscriber::FmtSubscriber;
    //
    //fn setup() {
    //    let lvl = Level::WARN;
    //    let subscriber = FmtSubscriber::builder().with_max_level(lvl).finish();
    //    tracing::subscriber::set_global_default(subscriber).unwrap();
    //}

    #[test]
    #[allow(clippy::too_many_lines)]
    fn group_record_tryfrom_homed_json_test() {
        //setup();

        let j = json!({
            "groupName" : "grobie",
            "binding" : {
                "6b18704270e94aa896b003b4340978f1" : {
                    "gid" : 60232
                }
            },
            "disposition" : "regular",
            "status" : {
                "6b18704270e94aa896b003b4340978f1" : {
                    "service" : "io.systemd.Home"
                }
            }
        });

        let g: GroupRecord = j.try_into().unwrap();

        // regular section
        assert_eq!(g.group_name, "grobie");
        assert!(matches!(g.disposition, GroupDisposition::GroupRegular));

        // binding section
        assert!(g.binding.is_some());
        let b = g.binding.unwrap();
        assert!(b.contains_key("6b18704270e94aa896b003b4340978f1"));
        let bk = b.get("6b18704270e94aa896b003b4340978f1").unwrap();
        assert_eq!(bk.gid, Some(60232));

        // status section
        assert!(g.status.is_some());
        let st = g.status.unwrap();
        assert!(st.contains_key("6b18704270e94aa896b003b4340978f1"));
        let stk = st.get("6b18704270e94aa896b003b4340978f1").unwrap();
        assert_eq!(stk.service, Some("io.systemd.Home".to_string()));
    }
}
