/*
 * Copyright (c) 2025      Andreas Schneider <asn@samba.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

//! This crate provides an API to access the systemd userdb API

#[cfg(feature = "capi")]
mod c_api;
#[cfg(feature = "capi")]
mod c_user_record;
#[cfg(feature = "capi")]
mod c_utils;
mod error;
mod group_record;
mod membership;
mod user_record;
pub use group_record::*;
use io_systemd_UserDatabase::VarlinkClientInterface;
pub use membership::*;
use tracing::error;
pub use user_record::*;
use varlink::Connection;
// This is auto-generated code
#[allow(
    dead_code,
    clippy::needless_lifetimes,
    clippy::redundant_closure_for_method_calls,
    clippy::uninlined_format_args
)]
mod io_systemd_UserDatabase;

pub use self::error::{Error, Result};

const SYSTEMD_MULIPLEXER_ADDRESS: &str = "unix:/run/systemd/userdb/io.systemd.Multiplexer";

/// The method looks up or enumerates a user record.
///
/// * `uid`: Specifies the numeric UNIX UID to search for.
/// * `user_name`: Specifies the name of the user to search for.
/// * `service`: Mandatory argument that should be set to the service name being talked to.
///              If you don't want to talk to a specific service, use: `io.systemd.Multiplexer`
///
/// # Returns
/// A vector of [`kirmes::UserRecord`][UserRecord]
///
/// # Errors
/// Returns [`kirmes::Error`][Error] if e.g. the connection fails or a record can't be parsed.
pub fn get_user_record(
    uid: Option<u32>,
    user_name: Option<String>,
    service: String,
) -> Result<Vec<UserRecord>> {
    let connection = match Connection::with_address(SYSTEMD_MULIPLEXER_ADDRESS) {
        Ok(c) => c,
        Err(e) => {
            error!("Failed to connect to {SYSTEMD_MULIPLEXER_ADDRESS}: {e:?}");
            return Err(Error::ServiceNotAvailable);
        },
    };

    let mut userdb_iface = io_systemd_UserDatabase::VarlinkClient::new(connection);

    let mut get_user_record = userdb_iface.get_user_record(uid, user_name, service);
    let mut ret = vec![];
    for reply in get_user_record.more()? {
        let reply = reply?;

        let record = reply.record;

        ret.push(record.try_into()?);
    }

    Ok(ret)
}

/// The method looks up or enumerates a group record.
///
/// * `gid`: Specifies the numeric UNIX GID to search for.
/// * `group_name`: Specifies the name of the group to search for.
/// * `service`: Mandatory argument that should be set to the service name being talked to.
///              If you don't want to talk to a specific service, use: `io.systemd.Multiplexer`
///
/// # Returns
/// A vector of [`kirmes::GroupRecord`][GroupRecord]
///
/// # Errors
/// Returns [`kirmes::Error`][Error] if e.g. the connection fails or a record can't be parsed.
pub fn get_group_record(
    gid: Option<u32>,
    group_name: Option<String>,
    service: String,
) -> Result<Vec<GroupRecord>> {
    let connection = match Connection::with_address(SYSTEMD_MULIPLEXER_ADDRESS) {
        Ok(c) => c,
        Err(e) => {
            error!("Failed to connect to {SYSTEMD_MULIPLEXER_ADDRESS}: {e:?}");
            return Err(Error::ServiceNotAvailable);
        },
    };

    let mut userdb_iface = io_systemd_UserDatabase::VarlinkClient::new(connection);

    let mut get_group_record = userdb_iface.get_group_record(gid, group_name, service);
    let mut ret: Vec<GroupRecord> = vec![];
    for reply in get_group_record.more()? {
        let reply = reply?;

        let record = reply.record;

        ret.push(record.try_into()?);
    }

    Ok(ret)
}

/// The method call may be used to inquire about group memberships.
///
/// The `user_name` and `group_name` arguments take what the name suggests. If one of the two is
/// specified all matching memberships are returned, if neither is specified all known memberships
/// of any user and any group are returned. The return value is a pair of user name and group name,
/// where the user is a member of the group.
///
/// * `user_name`: The user name to look up.
/// * `group_name`: The group name to look up.
/// * `service`: Mandatory argument that should be set to the service name being talked to.
///              If you don't want to talk to a specific service, use: `io.systemd.Multiplexer`
///
/// # Returns
/// A vector of [`kirmes::Membership`][Membership]
///
/// # Errors
/// Returns [`kirmes::Error`][Error] if e.g. the connection fails or a record can't be found.
pub fn get_memberships(
    user_name: Option<String>,
    group_name: Option<String>,
    service: String,
) -> Result<Vec<Membership>> {
    let connection = match Connection::with_address(SYSTEMD_MULIPLEXER_ADDRESS) {
        Ok(c) => c,
        Err(e) => {
            error!("Failed to connect to {SYSTEMD_MULIPLEXER_ADDRESS}: {e:?}");
            return Err(Error::ServiceNotAvailable);
        },
    };

    let mut userdb_iface = io_systemd_UserDatabase::VarlinkClient::new(connection);

    let mut get_memberships = userdb_iface.get_memberships(user_name, group_name, service);
    let mut ret: Vec<Membership> = vec![];
    for reply in get_memberships.more()? {
        let reply = reply?;

        let m = Membership {
            user_name: reply.userName,
            group_name: reply.groupName,
        };

        ret.push(m);
    }

    Ok(ret)
}
